/******************************************************************************

	COPYRIGHT(C) JONAS 'SORTIE' TERMANSEN 2011.

	This file is part of Sortix.

	Sortix is free software: you can redistribute it and/or modify it under the
	terms of the GNU General Public License as published by the Free Software
	Foundation, either version 3 of the License, or (at your option) any later
	version.

	Sortix is distributed in the hope that it will be useful, but WITHOUT ANY
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
	FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
	details.

	You should have received a copy of the GNU General Public License along
	with Sortix. If not, see <http://www.gnu.org/licenses/>.

	kernel.cpp
	This is where our kernel begins.

******************************************************************************/

#include "multiboot.h"
#include "types.h"
#include "itextscreen.h"
#include "vidmemtextscreen.h"
#include "iprintable.h"
#include "kerneloutput.h"
#include "panic.h"
#include "pageallocator.h"
#include "paging.h"
#include "gdt.h"
#include "idt.h"
#include "isr.h"
#include "memory.h"
#include "kernel.h"

// TODO: Move these to their own header!
inline void *operator new(size_t, void *p)     { return p; }
inline void *operator new[](size_t, void *p)   { return p; }
inline void  operator delete  (void *, void *) { };
inline void  operator delete[](void *, void *) { };

// We need a few global objects before the memory allocation is set up (used by
// the memory allocator, panic functions, and such). These objects are
// initialized in the KernelConstructors function.
#define DECLARE_GLOBAL_OBJECT(Type, Name) Type D##Name; Type* G##Name = &D##Name;
DECLARE_GLOBAL_OBJECT(KernelOutput, Out);
DECLARE_GLOBAL_OBJECT(PageAllocator, PageAllocator);
DECLARE_GLOBAL_OBJECT(Paging, Paging);
DECLARE_GLOBAL_OBJECT(GDT, GDT);
DECLARE_GLOBAL_OBJECT(IDT, IDT);
DECLARE_GLOBAL_OBJECT(ISR, ISR);

void KernelRun()
{
	GOut->SetColor(COLOR8_LIGHT_GREY);
	GOut->Print("(C) 2011 JONAS TERMANSEN                               ");

	GOut->SetColor(COLOR8_LIGHT_BROWN);
	GOut->Print("_");

	GOut->SetColor(COLOR8_LIGHT_GREY);
	GOut->Print("               GPLv3+  \n");
	
	GOut->SetColor(COLOR8_LIGHT_BROWN);
	GOut->Print("                                                      / \\                      \n");
	GOut->Print("                  /\\    /\\                           /   \\                     \n");
	GOut->Print("                 /  \\  /  \\                          |   |                     \n");
	GOut->Print("                /    \\/    \\                         |   |                     \n");
	GOut->Print("               |  O    O    \\_______________________ /   |                     \n");
	GOut->Print("               |                                         |                     \n");
	GOut->Print("               | \\_______/                               /                     \n");
	GOut->Print("                \\                                       /                      \n");
	GOut->Print("                  ------       ---------------      ---/                       \n");
	GOut->Print("                       /       \\             /      \\                          \n");
	GOut->Print("                      /         \\           /        \\                         \n");
	GOut->Print("                     /           \\         /          \\                        \n");
	GOut->Print("                    /_____________\\       /____________\\                       \n");
	GOut->Print("                                                                               \n");
	GOut->Print("                                                                               \n");
	
	GOut->SetColor(COLOR8_LIGHT_BROWN);
	GOut->Print("                           SORTIX KERNEL VERSION 1                             \n");

	GOut->SetColor(COLOR8_LIGHT_GREY);	
	GOut->Print("                                                                               \n");
	GOut->Print("                                                                               \n");
	GOut->Print("                                                                               \n");
	GOut->Print("                          \"Riverside, Motherfucker.\"                           \n");
	GOut->Print("                                                                               \n");
	GOut->Print("                                                                               \n");

	int b = 0; for ( size_t I = 0; I < 914891469; I++ ) { b += I; }

	const char* Messages[16] = {
		"Welcome to the Sortix Kernel\n",
		"This is the Sortix Kernel\n",
		"Welcome!\n",
		"You can do anything using the Sortix Kernel\n",

		"The only limit is yourself!\n",
		"Welcome to Sortix!\n",
		"Welcome!\n",
		"Anything is possible using Sortix!\n",

		"You can do anything with Sortix!\n",
		"The impossible is possible using the Sortix Kernel!\n",
		"In soviet russia, everything is possible using Sortix!\n",
		"This is a limitness world!\n",

		"Sortix can do more than Chuck Norris!\n",
		"Welcome to you who has come to Sortix!\n",
		"The infinite is possible using Sortix!\n",
		"The unattainable is unknown at Sortix!\n",
	};

	int Msg = 0;
	clrchar Color = 0;
	
	while ( true )
	{
		for ( size_t Msg = 0; Msg < 16; Msg++ )
		{
			if ( Color / 16 == Color % 16 ) { Color++; }

			int b = 0; for ( size_t I = 0; I < 114891469; I++ ) { b += I; }

			GOut->SetColor(Color);
			GOut->Print(Messages[Msg]);

			GOut->Flush();

			Color = (Color + 1) % (COLOR8_LIGHT_GREY<<4);
		}
	}
}

void KernelConstructors()
{
	GOut = new (GOut) KernelOutput;
	GPageAllocator = new (GPageAllocator) PageAllocator;
	GPaging = new (GPaging) Paging;
	GGDT = new (GGDT) GDT;
	GIDT = new (GIDT) IDT;
	GISR = new (GISR) ISR;
}

#define TO_MIB(Bytes) ( (uint64_t) (Bytes / 1024ULL / 1024ULL) )

void ParseMultiboot(unsigned long Magic, multiboot_info_t* BootInfo)
{
	if ( !BootInfo )
	{
		Panic("kernel.cpp: The bootinfo structure was NULL. Are your bootloader multiboot compliant?");
	}

	if ( !( BootInfo->flags & (1<<6) ) )
	{
		Panic("kernel.cpp: The memory map flag was't set in the multiboot structure. Are your bootloader multiboot compliant?");
	}

	GOut->Print("Start of multiboot debug info\n");

	GOut->PrintF("mmap_addr = 0x%x, mmap_length = 0x%x\n", BootInfo->mmap_addr, BootInfo->mmap_length);

	uint64_t Available = 0;
	uint64_t Reserved = 0;
	uint64_t Unaddressable = 0;
	
    for	(
			multiboot_memory_map_t* mmap = (multiboot_memory_map_t*) BootInfo->mmap_addr;
			(unsigned long) mmap < BootInfo->mmap_addr + BootInfo->mmap_length;
			mmap = (multiboot_memory_map_t *) ((unsigned long) mmap + mmap->size + sizeof(mmap->size))
		)
	{
		bool Addressable = true;
#ifdef PLATFORM_X86
		// Figure out if the memory area is addressable (are our pointers big enough?)
		if ( 0xFFFFFFFFULL < mmap->addr || 0xFFFFFFFFULL < mmap->addr + mmap->len )
		{
			Addressable = false;
		}
#endif

		if ( Addressable )
		{
			if ( mmap->type == 1 ) { Available += mmap->len; GOut->SetColor(COLOR8_GREEN); } else { Reserved += mmap->len; GOut->SetColor(COLOR8_RED); }
		}
		else
		{
			Unaddressable += mmap->len;
			GOut->SetColor(COLOR8_BLUE);
		}

		GOut->PrintF(" base_addr = 0x%64x, length = 0x%64x\n",
			mmap->addr,
			mmap->len);

	}

	GOut->SetColor(COLOR8_LIGHT_GREY);

	GOut->PrintF("RAM: Available %64u MiB, Reserved %64u MiB, Unaddressable %64u MiB, Total %64u MiB\n", TO_MIB(Available), TO_MIB(Reserved), TO_MIB(Unaddressable), TO_MIB(Unaddressable)+TO_MIB(Available)+TO_MIB(Reserved));

	GOut->Print("End of multiboot debug info\n");

	//HaltKernel();

	//for ( int I = 10; I > 0; I-- ) { GOut->PrintF("\rProceeding with boot in %u seconds...   ", I); GOut->Flush(); int b = 0; for ( size_t I = 0; I < 314891469; I++ ) { b += I; } } GOut->Print("\n");
}

void HackSleep(size_t Miliseconds)
{
	//int b = 0; for ( size_t I = 0; I < (5148 * Miliseconds); I++ ) { b += I; }
	int b = 0; for ( size_t I = 0; I < (314891 * Miliseconds); I++ ) { b += I; }
}

extern "C" void KernelInit(unsigned long Magic, multiboot_info_t* BootInfo)
{
	// Initialize global objects manually, since globals objects never have
	// their constructors called. (We are a kernel, who would do it but us?)
	KernelConstructors();

	// Alright, we have now set up our environment, it's time to read the
	// multiboot data provided to us by GRUB.
	ParseMultiboot(Magic, BootInfo);

	// Initialize the Global Descriptor Table (IDT) to handle something about memory access and segmentation.
	GGDT->Init();

	// Initialize the Interrupt Descriptor Table (IDT) to handle interrupts.
	GIDT->Init();

	int b = 5 / 0;

	Panic("SUCCESS!");

	// Next up, memory allocating needs to be set up. Until now our kernel has
	// just used static memory, but it's time to utilize the many gigabytes our
	// user has been so kind to provide us with. Our multiboot compliant boot-
	// loader has provided us with useful information about how the physical
	// memory has been arranged, so let's utilize that. Also tell our page
	// allocator that the physical pages 0x100000 and 0x100000 bytes forward
	// are reserved for use by the Sortix kernel. (We wouldn't like very much
	// if a user was watching a funny cat video on Youtube and it somehow
	// ended up inside our kernel machine code.)
	// TODO: Kids probably won't even have heard of Youtube when this kernel
	// is finally released.
	GPageAllocator->Init(BootInfo, 0x100000, 0x100000);

	// Now that our physical memory have been set up, let's set up our virtual
	// memory (using paging). This is a pretty cool feature.
	GPaging->Init(0x100000, 0x100000);

	// TODO: Initialize the kernel memory allocator here!

	// Alright, our kernel has been initialized. It's time to actually start
	// the mainloop and let everything run.
	KernelRun();
}


