/******************************************************************************

	COPYRIGHT(C) JONAS 'SORTIE' TERMANSEN 2011.

	This file is part of Sortix.

	Sortix is free software: you can redistribute it and/or modify it under the
	terms of the GNU General Public License as published by the Free Software
	Foundation, either version 3 of the License, or (at your option) any later
	version.

	Sortix is distributed in the hope that it will be useful, but WITHOUT ANY
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
	FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
	details.

	You should have received a copy of the GNU General Public License along
	with Sortix. If not, see <http://www.gnu.org/licenses/>.

	pageallocator.cpp
	Allocates physical memory pages for use by the system.

******************************************************************************/

#include "multiboot.h"
#include "types.h"
#include "iprintable.h"
#include "itextscreen.h"
#include "vidmemtextscreen.h"
#include "kerneloutput.h"
#include "pageallocator.h"
#include "panic.h"
#include "kernel.h"

#define InlineAllocateAddress(Address) InlineAllocatePage(0, ((size_t) Address)/4096);

#ifdef PLATFORM_X86
#define FOR_EACH_MEMORY_AREA_IN_ADDRESSABLE \
		/* Figure out if the memory area is addressable (are our pointers big enough?) */ \
		if ( 0xFFFFFFFFULL < mmap->addr || 0xFFFFFFFFULL < mmap->addr + mmap->len ) { Addressable = false; }
#else
#define FOR_EACH_MEMORY_AREA_IN_ADDRESSABLE
#endif

#define FOR_EACH_MEMORY_AREA_IN(BootInfo) \
    for	( \
			multiboot_memory_map_t* mmap = (multiboot_memory_map_t*) BootInfo->mmap_addr; \
			(unsigned long) mmap < BootInfo->mmap_addr + BootInfo->mmap_length; \
			mmap = (multiboot_memory_map_t *) ((unsigned long) mmap + mmap->size + sizeof(mmap->size)) \
		) \
	{ \
		bool Addressable = true; \
		FOR_EACH_MEMORY_AREA_IN_ADDRESSABLE

void PageAllocator::Init(multiboot_info_t* BootInfo, size_t KernelStart, size_t KernelSize)
{
	// First, let's do some sanity checking!
	if ( !BootInfo ) { Panic("kernel.cpp: The bootinfo structure was NULL. Are your bootloader multiboot compliant?"); }
	if ( !( BootInfo->flags & (1<<6) ) ) { Panic("kernel.cpp: The memory map flag was't set in the multiboot structure. Are your bootloader multiboot compliant?"); }

	// Now we need to measure how much physical memory we have available.
	MeasureRam(BootInfo);
	// Since we know the amount of physical available memory now, we have to
	// start book-keeping which pages have been allocated. This involves a
	// rather big bitmap, but it should be of considerable performance.
	AllocatePageBitmaps(BootInfo, KernelStart, KernelSize);
}

void PageAllocator::MeasureRam(multiboot_info_t* BootInfo)
{
	// Reset our counters of system resources.
	AvailableRam			=	0;
	AvailableEnd 			=	0;
	ReservedRam				=	0;
	UnaddressableRam		=	0;

	// First, let's traverse the multiboot structure and measure how much RAM we have available.
   	FOR_EACH_MEMORY_AREA_IN(BootInfo)
//	{
		if ( mmap->type == 1 )
		{
			if ( Addressable )
			{
				AvailableRam += mmap->len;
				if ( AvailableEnd < mmap->addr + mmap->len ) { AvailableEnd = mmap->addr + mmap->len; }
			}
			else
			{
				UnaddressableRam += mmap->len;
			}
		}
		else
		{
			ReservedRam += mmap->len;
		}
	}

	TotalRam = AvailableRam + ReservedRam + UnaddressableRam;

	GOut->PrintF("Found total ram %64u!\n", TotalRam);
}

#define DIVROUNDUP(A, B) ( (A + B - 1) / B )
#define ALIGNUPTO(A, B) ( DIVROUNDUP(A, B) * B )

void PageAllocator::AllocatePageBitmaps(multiboot_info_t* BootInfo, size_t KernelStart, size_t KernelSize)
{
	// How many pages shall we handle?
	NumPages = AvailableEnd / 4096;

	// Let's find out how many layered bitmaps we need, and how much memory it adds up to.
	size_t BufferSize = 0;
	size_t LayerEntries = NumPages;
	BitmapLayers = 0;

	while ( LayerEntries > 0 )
	{
		BufferSize += ALIGNUPTO(DIVROUNDUP(LayerEntries, 8), sizeof(size_t));
		BitmapLayers++;
		
		if ( LayerEntries == 1 ) { break; }
		LayerEntries = DIVROUNDUP(LayerEntries, 2);
	}

	GOut->PrintF("Using %z bitmap layers!\n", BitmapLayers);

	// Allocate room for some pointers and widths too.
	BufferSize += BitmapLayers * sizeof(size_t) * 3;

	// We now need to allocate a bitmap of which pages are available, and of
	// which pages aren't. Of course, this is complicated by the fact that
	// there is no other memory allocator, so we have to figure out ourselves
	// where to put it. Luckily, virtually all the memory is unused except for
	// the kernel and other reserved areas. So let's find a region large enough
	// to hold our bitmap.
	// TODO: We might want to have set up virtual memory before this?

	bool FoundRegion = false;
	void* Buffer;

   	FOR_EACH_MEMORY_AREA_IN(BootInfo)
//	{
		if ( Addressable && mmap->type == 1 )
		{
			// Check if the kernel overlaps with this memory region.
			if ( mmap->addr + mmap->len < KernelStart || KernelStart + KernelSize < mmap->addr )
			{
				if ( BufferSize <= mmap->len )
				{
					FoundRegion = true; Buffer = (void*) mmap->addr; break;
				}
			}
			else
			{
				// Looks like our kernel is in this region too, let's see if we
				// have space beside it.
				if ( BufferSize < KernelStart - mmap->addr )
				{
					FoundRegion = true; Buffer = (void*) mmap->addr; break;
				}
				else if ( BufferSize < mmap->len - KernelSize )
				{
					FoundRegion = true; Buffer = (void*) (mmap->addr + KernelSize); break;
				}
			}
		}
	}

	if ( !FoundRegion ) { Panic("pageallocator.cpp: Could not find room for the page allocation bitmaps!"); }

	// Alright, we got ourselves a memory region!
	GOut->PrintF("Allocated region at 0x%p spanning %z KiB!\n", Buffer, BufferSize / 1024);
	GOut->Flush();

	// Alright, let's reset the memory area we found.
	BitmapsSize = BufferSize;
	size_t BitmapLength = BitmapsSize/sizeof(size_t);
	Bitmaps = (size_t**) Buffer; for ( size_t I = 0; I < BitmapLength; I++ ) { ((size_t*)Bitmaps)[I] = 0; }
	BitmapWidths = ((size_t*)Bitmaps) + 1 * BitmapLayers;
	BitmapEntries = ((size_t*)Bitmaps) + 2 * BitmapLayers;

	// Now let's set up our pointers.
	char* Location  = (char*) ( ((size_t*)Bitmaps) + 3 * BitmapLayers ); // Remember to skip the list of meta deta.

	// Set up a pointer for each bitmap.
	size_t Entries = NumPages;
	for ( size_t I = 0; I < BitmapLayers; I++, Entries = DIVROUNDUP(Entries, 2) )
	{
		size_t BitmapWidth = ALIGNUPTO(DIVROUNDUP(Entries, 8), sizeof(size_t));

		//GOut->PrintF("(%z) Entries=%z ==> width=%z\t", I, Entries, BitmapWidth);

		Bitmaps[I] = (size_t*) Location;
		BitmapWidths[I] = BitmapWidth;
		BitmapEntries[I] = Entries;
		//Bitmaps[I][0] = 0xFACE;
		//Bitmaps[I][1] = 0xB00C;
		Location += BitmapWidth;
	}
	GOut->Print("\n");

	// Reset our counters!
	NumUsedPages = 0;
	NumFreePages = NumPages;

	// Now allocate our own table.
	for ( char* I = (char*) Bitmaps; I < (char*) Bitmaps + BitmapsSize; I += 4096 ) { InlineAllocateAddress(I) }

	// Now allocate our kernel.
	for ( size_t I = KernelStart; I < KernelStart + KernelSize; I += 4096 ) { InlineAllocateAddress(I) }

	// Now allocate all the reserved memory areas.
   	FOR_EACH_MEMORY_AREA_IN(BootInfo)
//	{
		if ( Addressable && mmap->type != 1 && mmap->addr < AvailableEnd )
		{
			for ( uint64_t I = mmap->addr; I < mmap->addr + mmap->len; I += 4096 ) { InlineAllocateAddress(I) }
		}
	}

	//GOut->PrintF("Allocated %z pages, %z pages free\n", NumUsedPages, NumFreePages);
	//GOut->PrintF("Allocated %z MiB, %z MiB free\n", NumUsedPages / 256, NumFreePages / 256);

	size_t Allocated = 0;
	void* Page;
	void* LastPage = NULL;
	GOut->Print("Performing memory allocation sanity test...\n");
	GOut->SetColor(COLOR8_DARK_GREY << 4); for ( size_t I = 0; I < GOut->GetWidth(); I++ ) { GOut->Print(" "); } GOut->Print("\r"); GOut->SetColor(COLOR8_LIGHT_GREY << 4); GOut->Flush();
	size_t Progress = 0;
	
	while ( (Page = AllocatePage()) != NULL )
	{
		//if ( LastPage == Page ) { GOut->PrintF("Allocated same page %p twice!\n", Page); HaltKernel(); }
		Allocated++;
		//GOut->PrintF("\r%z MiB", Page, Allocated / 256);
		for ( ; Progress < NumUsedPages * GOut->GetWidth() / NumPages; Progress++ ) { GOut->Print(" "); FreePage(Page); }
		GOut->Flush();
		LastPage = Page;

		for ( size_t I = 0; I < 1000; I++ )
		{
			((size_t*) Page)[I] = 0xFACEB00C;
		}
	}
	GOut->Print("\n");

	GOut->PrintF("Allocated %z pages, %z pages free\n", NumUsedPages, NumFreePages);
	GOut->PrintF("Allocated %z MiB, %z MiB free\n", NumUsedPages / 256, NumFreePages / 256);

	HaltKernel();
}

inline void PageAllocator::InlineAllocatePage(size_t BitmapLayer, size_t BitNum)
{
	if ( BitmapLayer == 0 ) { NumUsedPages++; NumFreePages--; }

LayerUp:

	size_t Offset = BitNum / (8*sizeof(size_t));
	size_t Bit = BitNum % (8*sizeof(size_t));

	Bitmaps[BitmapLayer][Offset] = Bitmaps[BitmapLayer][Offset] | (1<<Bit);

	// Are we the root bitmap?
	if ( BitmapLayer == BitmapLayers-1 ) { return; }

	// Check if its sibling is allocated. If so, then mark out parent as being allocated.
	size_t SiblingBit = Bit ^ 1;

	// If we are an only child, then mark our parent as allocated.
	if ( BitmapEntries[BitmapLayer] - 1 == BitNum )
	{
		// Be sure we don't have a sibling!
		if ( BitNum % 2 == 0 ) { BitmapLayer++; BitNum /= 2; goto LayerUp; }
	}

	if ( Bitmaps[BitmapLayer][Offset] & (1<<SiblingBit) )
	{
		// Alright, if I, and my sibling are allocated, then our parent is too!
		BitmapLayer++; BitNum /= 2; goto LayerUp;
	}
}

// Finds an unallocated page in O(log n)
// TODO: When 0xCFEFF000 is the end of the allocatable memory, the function also allocates 0xCFEFF000, which is horribly wrong. Fix that!
void* PageAllocator::AllocatePage()
{
	size_t Layer = BitmapLayers - 1;
	
	// Check the root layer to see if we are out of memory. 
	if ( Bitmaps[Layer][0] & 0x1 ) { return NULL; }

	size_t BitNum = 0;

	// Now recurse until we find an unallocated page.
	for ( Layer = Layer; Layer > 0; Layer-- )
	{
		size_t Offset = (2*BitNum) / (8*sizeof(size_t));
		size_t Word = Bitmaps[Layer-1][Offset];
		size_t Bit1 = (2*BitNum+0) % (8*sizeof(size_t));
		size_t Bit2 = (2*BitNum+1) % (8*sizeof(size_t));

		//GOut->PrintF("%p + %z (%z): word=%x, bit1=%u>%u>%x, bit2=%u>%u>%x\n", Bitmaps[Layer-1], Offset, BitNum, Word,
		//		Bit1, (1<<Bit1), Word & (1<<Bit1), 
		//		Bit2, (1<<Bit2), Word & (1<<Bit2));
		if ( (Word & (1<<Bit1)) == 0 )
		{
			BitNum = BitNum * 2 + 0;
		}
		else if ( (Word & (1<<Bit2)) == 0 )
		{
			// In base Bit1 is an only child, then this bit should never be
			// touched as its value is undefined. However, if the
			// InlineAllocatePage function performs properly, then if an only
			// child is allocated, so is its parent, and thus we would never
			// run this code.
			BitNum = BitNum * 2 + 1;
			if ( BitmapEntries[Layer-1] <= BitNum ) { GOut->Print("pageallocator.cpp: Touched naughty bit!\n"); HaltKernel(); }
		}
		else
		{
			GOut->PrintF("Invariant broken for word %x in layer %z, where both bit %u and %u is set!\n", Word, Layer-1, Bit1, Bit2);
			GOut->PrintF("%p + %z (%z): ", Bitmaps[Layer-1], Offset, BitNum);
			for ( int I = -80; I < 80; I++ ) { GOut->SetColor((I==0) ? COLOR8_RED : COLOR8_LIGHT_GREY); GOut->PrintF("%x ", Bitmaps[Layer-1][Offset + I]); } GOut->SetColor(COLOR8_RED); GOut->Print("\n");
			HaltKernel();
		}
	}

	// Alright, now mark the page as allocated. Runs in O(log n) in worst time,
	// but runs in O(1) in best time. It's fast enough.
	InlineAllocatePage(0, BitNum);

	return (char*) (BitNum * 4096);
}

void PageAllocator::FreePage(void* Page)
{
	if ( (size_t)Page % 0x1000 != 0 )
	{
		GOut->PrintF("Expected page input %p to be page aligned (multiple of %x)\n", Page, (uint32_t) 0x1000);
		HaltKernel();
	}

	size_t BitNum = ((size_t)Page) / 4096;
	
	InlineFreePage(0, BitNum);	
}

inline void PageAllocator::InlineFreePage(size_t BitmapLayer, size_t BitNum)
{
	if ( BitmapLayer == 0 ) { NumUsedPages--; NumFreePages++; }

	size_t Offset = BitNum / (8*sizeof(size_t));
	size_t Bit = BitNum % (8*sizeof(size_t));

	Bitmaps[BitmapLayer][Offset] = Bitmaps[BitmapLayer][Offset] & ~(1<<Bit);

	// Are we the root bitmap?
	if ( BitmapLayer == BitmapLayers-1 ) { return; }

	// Check if its sibling is deallocated. If so, then mark out parent as being deallocated.
	size_t SiblingBit = Bit ^ 1;

	// If we are an only child, then mark our parent as deallocaed.
	if ( BitmapEntries[BitmapLayer] - 1 == BitNum )
	{
		// Be sure we don't have a sibling!
		if ( BitNum % 2 == 0 ) { InlineFreePage(BitmapLayer + 1, BitNum / 2); return; }
	}

	if ( !(Bitmaps[BitmapLayer][Offset] & (1<<SiblingBit)) )
	{
		// Alright, if I, and my sibling are deallocated, then our parent is too!
		InlineFreePage(BitmapLayer + 1, BitNum / 2);
	}
}

void PageAllocator::IdentityMap()
{
	// TODO: Implement this!
}

