/******************************************************************************

	COPYRIGHT(C) JONAS 'SORTIE' TERMANSEN 2011.

	This file is part of Sortix.

	Sortix is free software: you can redistribute it and/or modify it under the
	terms of the GNU General Public License as published by the Free Software
	Foundation, either version 3 of the License, or (at your option) any later
	version.

	Sortix is distributed in the hope that it will be useful, but WITHOUT ANY
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
	FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
	details.

	You should have received a copy of the GNU General Public License along
	with Sortix. If not, see <http://www.gnu.org/licenses/>.

	kerneloutput.cpp
	Lets the kernel print possibly useful messages.

******************************************************************************/

#include "types.h"
#include "itextscreen.h"
#include "vidmemtextscreen.h"
#include "iprintable.h"
#include "kerneloutput.h"
#include "panic.h"

KernelOutput::KernelOutput()
{
	Constructed = 1337;

	Width = 80;
	Height = 25;
	Color = (COLOR8_BLACK << 12) + (COLOR8_LIGHT_GREY << 8);
	Pos = 0;

	// TODO: A memset would be faster, but we haven't implemented that yet.
	// The compiler will probably optimize this regardless.
	for ( size_t I = 0; I < Width*Height; I++ )
	{
		Text[I] = ' ' | Color;
	}

	Flush();
}


clrchar KernelOutput::SetColor(clrchar NewColor)
{
	clrchar OldColor = Color;
	Color = NewColor << 8;
	return OldColor;
}

void KernelOutput::Print(const char* String)
{
	if ( Constructed != 1337 ) { Panic("kerneloutput.cpp: Attempted to use non-constructed kerneloutput"); }

	while ( *String != '\0' )
	{
		if ( *String == '\n' )
		{
			Newline();
		}
		else if ( Pos % Width == Width -1 )
		{
			Newline();
			Text[Pos] = *String | Color; Pos++;
		}
		else if ( *String == '\r' )
		{
			Pos = (Pos/Width) * Width;
		}
		else if ( *String == '\t' )
		{
			if ( Pos % Width >= Width-1 - 4 )
			{
				for ( ; Pos % Width != 0; Pos++ ) { Text[Pos] = ' ' | Color; }
				Pos--;
				Newline();
			}
			else
			{
				size_t TabEnd = ((Pos+4)/4)*4;
				for ( ; Pos < TabEnd; Pos++ ) { Text[Pos] = ' ' | Color; }
			}
		}
		else
		{
			Text[Pos] = *String | Color; Pos++;
		}

		String++;
	}
}

void KernelOutput::Newline()
{
	if ( Pos / Width < Height - 1 )
	{
		Pos = (Pos/Width + 1) * Width;
	}
	else
	{
		for ( size_t I = 1; I < Height; I++ )
		{
			for ( size_t N = 0; N < Width; N++ )
			{
				Text[(I-1) * Width + N] = Text[(I) * Width + N];
			}
		}

		Pos = (Pos/Width) * Width;

		for ( size_t I = 0; I < Width; I++ )
		{
			Text[Pos + I] = ' ' | Color;
		}
	}

	Flush();
}

void KernelOutput::Flush()
{
	Screen.Seek(0, 0);
	Screen.WriteRaw(Text, Width*Height * sizeof(clrchar));
}


void KernelOutput::Debug(const char* String)
{
	clrchar Old = SetColor(COLOR8_GREEN);
	Print(String);
	SetColor(Old);
}

void KernelOutput::Error(const char* String)
{
	clrchar Old = SetColor(COLOR8_RED);
	Print(String);
	SetColor(Old);
}

void KernelOutput::Log(const char* String)
{
	clrchar Old = SetColor(COLOR8_LIGHT_GREY);
	Print(String);
	SetColor(Old);
}

void KernelOutput::Warning(const char* String)
{
	clrchar Old = SetColor(COLOR8_CYAN);
	Print(String);
	SetColor(Old);
}

